"""Main CLI entry point for Opik."""

from importlib import metadata

import click

from .configure import configure
from .export import export
from .healthcheck import healthcheck
from .import_command import import_data
from .proxy import proxy

__version__: str = "0.0.0+dev"
try:
    __version__ = metadata.version("opik")
except metadata.PackageNotFoundError:
    pass

CONTEXT_SETTINGS = {"help_option_names": ["-h", "--help"]}


@click.group(invoke_without_command=True, context_settings=CONTEXT_SETTINGS)
@click.version_option(__version__, *("--version", "-v"))
def cli() -> None:
    """CLI tool for Opik."""


# Register all commands
cli.add_command(configure)
cli.add_command(proxy)
cli.add_command(healthcheck)
cli.add_command(export)
cli.add_command(import_data)



import functools
import inspect
import logging
from typing import List, Optional, Callable, Any, Dict

import opik.exceptions as exceptions
import opik.logging_messages as logging_messages
import opik.opik_context as opik_context
import opik
from opik.api_objects import opik_client, trace
from opik.api_objects.dataset import dataset, dataset_item
from opik.api_objects.experiment import experiment
from opik.evaluation import (
    rest_operations,
    test_case,
    test_result,
    samplers,
)
from opik.evaluation.types import LLMTask, ScoringKeyMappingType

from . import evaluation_tasks_executor, exception_analyzer, helpers
from .types import EvaluationTask
from ..metrics import arguments_validator, arguments_helpers, base_metric, score_result
from ...message_processing import message_processors_chain
from ...message_processing.emulation import models


LOGGER = logging.getLogger(__name__)

EVALUATION_TASK_NAME = "evaluation_task"
EVALUATION_SPAN_PARAMETER_NAME = "task_span"


class EvaluationEngine:
    def __init__(
        self,
        client: opik_client.Opik,
        project_name: Optional[str],
        experiment_: experiment.Experiment,
        scoring_metrics: List[base_metric.BaseMetric],
        workers: int,
        verbose: int,
        scoring_key_mapping: Optional[ScoringKeyMappingType],
    ) -> None:
        self._client = client
        self._project_name = project_name
        self._experiment = experiment_
        self._workers = workers
        self._verbose = verbose
        self._scoring_metrics: List[base_metric.BaseMetric] = []
        self._task_span_scoring_metrics: List[base_metric.BaseMetric] = []
        self._scoring_key_mapping = scoring_key_mapping

        # Analyze metrics
        self._analyze_metrics(scoring_metrics)

        if len(self._task_span_scoring_metrics) > 0:
            LOGGER.info(
                "Detected %d LLM task span scoring metrics — enabling handling of the LLM task evaluation span.",
                len(self._task_span_scoring_metrics),
            )

    def _analyze_metrics(self, scoring_metrics: List[base_metric.BaseMetric]) -> None:
        for metric in scoring_metrics:
            if _has_evaluation_span_parameter(metric.score):
                self._task_span_scoring_metrics.append(metric)
            else:
                self._scoring_metrics.append(metric)

    @opik.track(name="metrics_calculation")  # type: ignore[attr-defined,has-type]
    def _evaluate_test_case(
        self,
        test_case_: test_case.TestCase,
        trial_id: int = 0,
    ) -> test_result.TestResult:
        score_results = _scores_by_metrics(
            scoring_metrics=self._scoring_metrics,
            score_kwargs=test_case_.scoring_inputs,
            scoring_key_mapping=self._scoring_key_mapping,
        )

        test_result_ = test_result.TestResult(
            test_case=test_case_,
            score_results=score_results,
            trial_id=trial_id,
        )
        rest_operations.log_test_result_feedback_scores(
            client=self._client,
            score_results=score_results,
            trace_id=test_case_.trace_id,
            project_name=self._project_name,
        )
        return test_result_

    def _evaluate_llm_task(
        self,
        item: dataset_item.DatasetItem,
        task: LLMTask,
        trial_id: int,
    ) -> test_result.TestResult:
        if not hasattr(task, "opik_tracked"):
            name = task.__name__ if hasattr(task, "__name__") else "llm_task"
            task = opik.track(name=name)(task)  # type: ignore[attr-defined,has-type]

        item_content = item.get_content(include_id=True)
        trace_data = trace.TraceData(
            input=item_content,
            name=EVALUATION_TASK_NAME,
            created_by="evaluation",
            project_name=self._project_name,
        )

        with helpers.evaluate_llm_task_context(
            experiment=self._experiment,
            dataset_item_id=item.id,
            trace_data=trace_data,
            client=self._client,
        ):
            LOGGER.debug("Task started, input: %s", item_content)
            try:
                task_output_ = task(item_content)
            except Exception as exception:
                if exception_analyzer.is_llm_provider_rate_limit_error(exception):
                    LOGGER.error(
                        logging_messages.LLM_PROVIDER_RATE_LIMIT_ERROR_DETECTED_IN_EVALUATE_FUNCTION
                    )

                raise
            LOGGER.debug("Task finished, output: %s", task_output_)

            opik_context.update_current_trace(output=task_output_)

            scoring_inputs = arguments_helpers.create_scoring_inputs(
                dataset_item=item_content,
                task_output=task_output_,
                scoring_key_mapping=self._scoring_key_mapping,
            )

            test_case_ = test_case.TestCase(
                trace_id=trace_data.id,
                dataset_item_id=item.id,
                scoring_inputs=scoring_inputs,
                task_output=task_output_,
            )
            test_result_ = self._evaluate_test_case(
                test_case_=test_case_,
                trial_id=trial_id,
            )

        return test_result_

    def evaluate_llm_tasks(
        self,
        dataset_: dataset.Dataset,
        task: LLMTask,
        nb_samples: Optional[int],
        dataset_item_ids: Optional[List[str]],
        dataset_sampler: Optional[samplers.BaseDatasetSampler],
        trial_count: int,
    ) -> List[test_result.TestResult]:
        task_span_scoring_enabled = False
        if len(self._task_span_scoring_metrics) > 0:
            message_processors_chain.toggle_local_emulator_message_processor(
                active=True, chain=self._client._message_processor
            )
            task_span_scoring_enabled = True

        dataset_items = dataset_.__internal_api__get_items_as_dataclasses__(
            nb_samples=nb_samples,
            dataset_item_ids=dataset_item_ids,
        )

        if dataset_sampler is not None:
            dataset_items = dataset_sampler.sample(dataset_items)

        test_results: List[test_result.TestResult] = []

        for trial_id in range(trial_count):
            evaluation_tasks: List[EvaluationTask[test_result.TestResult]] = [
                functools.partial(
                    self._evaluate_llm_task,
                    item=item,
                    task=task,
                    trial_id=trial_id,
                )
                for item in dataset_items
            ]

            test_results += evaluation_tasks_executor.execute(
                evaluation_tasks,
                self._workers,
                self._verbose,
                desc=f"Evaluation trial {trial_id}"
                if trial_count > 1
                else "Evaluation",
            )

        if task_span_scoring_enabled:
            # flush Opik client to make sure all spans are collected
            self._client.flush()

            self._evaluate_llm_tasks_spans(test_results)

            LOGGER.info(
                "Task evaluation span handling is disabled — the evaluation has been completed."
            )
            message_processors_chain.toggle_local_emulator_message_processor(
                active=False, chain=self._client._message_processor
            )

        return test_results

    def _evaluate_llm_tasks_spans(
        self, test_results: List[test_result.TestResult]
    ) -> None:
        local = message_processors_chain.get_local_emulator_message_processor(
            chain=self._client._message_processor
        )
        if local is None:
            LOGGER.warning("Local emulator message processor not found in the chain.")
            return

        # get trace trees from a local emulator
        trace_trees = local.trace_trees
        if len(trace_trees) == 0:
            LOGGER.warning("No trace trees found in the local emulator.")
            return

        # create span evaluation tasks from LLM tasks evaluation results and evaluate them in parallel
        span_evaluation_tasks: List[EvaluationTask[test_result.TestResult]] = [
            functools.partial(
                self._evaluate_llm_task_result_span,
                evaluation_task_result=test_result_,
                trace_trees=trace_trees,
            )
            for test_result_ in test_results
        ]

        evaluation_tasks_executor.execute(
            span_evaluation_tasks,
            self._workers,
            self._verbose,
            desc="LLM task spans evaluation",
        )

    def _evaluate_llm_task_result_span(
        self,
        evaluation_task_result: test_result.TestResult,
        trace_trees: List[models.TraceModel],
    ) -> test_result.TestResult:
        # find related trace
        trace_id = evaluation_task_result.test_case.trace_id
        task_trace = None
        for trace_ in trace_trees:
            if trace_.id == trace_id:
                task_trace = trace_
                break

        if task_trace is None:
            raise ValueError(
                f"No trace found for test result: {evaluation_task_result}"
            )

        # find evaluation span
        if len(task_trace.spans) == 0:
            raise ValueError(
                f"Task trace contains no spans. Task span metrics require at least one span to be present in the execution trace. Test result: {evaluation_task_result}"
            )
        # the first span is the evaluation span
        evaluation_span = task_trace.spans[0]

        with helpers.evaluate_llm_task_result_spans_context(
            trace_data=trace.TraceData(
                id=trace_id,
                name=task_trace.name,
                start_time=task_trace.start_time,
                metadata=task_trace.metadata,
                input=task_trace.input,
                output=task_trace.output,
                tags=task_trace.tags,
                project_name=self._project_name,
                created_by="evaluation",
                error_info=task_trace.error_info,
                thread_id=task_trace.thread_id,
            ),
            client=self._client,
        ):
            score_results = self._score_llm_task_result_span(
                trace_id=trace_id,
                task_span=evaluation_span,
                test_case_=evaluation_task_result.test_case,
            )
            # append scores to the input test result
            evaluation_task_result.score_results += score_results
            return evaluation_task_result

    @opik.track(  # type: ignore[attr-defined,has-type]
        name="task_span_metrics_calculation",
        ignore_arguments=["test_case_"],
    )
    def _score_llm_task_result_span(
        self,
        trace_id: str,
        task_span: models.SpanModel,
        test_case_: test_case.TestCase,
    ) -> List[score_result.ScoreResult]:
        score_kwargs = {
            **test_case_.scoring_inputs,
            EVALUATION_SPAN_PARAMETER_NAME: task_span,
        }

        score_results = _scores_by_metrics(
            scoring_metrics=self._task_span_scoring_metrics,
            score_kwargs=score_kwargs,
            scoring_key_mapping=self._scoring_key_mapping,
        )

        # log feedback scores
        rest_operations.log_test_result_feedback_scores(
            client=self._client,
            score_results=score_results,
            trace_id=trace_id,
            project_name=self._project_name,
        )
        return score_results

    def evaluate_test_cases(
        self,
        test_cases: List[test_case.TestCase],
    ) -> List[test_result.TestResult]:
        evaluation_tasks: List[EvaluationTask[test_result.TestResult]] = [
            functools.partial(
                self._evaluate_test_case,
                test_case_=test_case_,
            )
            for test_case_ in test_cases
        ]

        test_results = evaluation_tasks_executor.execute(
            evaluation_tasks, self._workers, self._verbose
        )

        return test_results


def _scores_by_metrics(
    scoring_metrics: List[base_metric.BaseMetric],
    score_kwargs: Dict[str, Any],
    scoring_key_mapping: Optional[ScoringKeyMappingType],
) -> List[score_result.ScoreResult]:
    score_results: List[score_result.ScoreResult] = []
    for metric in scoring_metrics:
        try:
            arguments_validator.validate_score_arguments(
                metric=metric,
                kwargs=score_kwargs,
                scoring_key_mapping=scoring_key_mapping,
            )
            LOGGER.debug("Metric %s score started", metric.name)
            result = metric.score(**score_kwargs)
            LOGGER.debug("Metric %s score ended", metric.name)

            if isinstance(result, list):
                score_results += result
            else:
                score_results.append(result)
        except exceptions.ScoreMethodMissingArguments:
            raise
        except Exception as exception:
            # This can be problematic if the metric returns a list of strings as we will not know the name of the metrics that have failed
            LOGGER.error(
                "Failed to compute metric %s. Score result will be marked as failed.",
                metric.name,
                exc_info=True,
            )

            if exception_analyzer.is_llm_provider_rate_limit_error(exception):
                LOGGER.error(
                    logging_messages.LLM_PROVIDER_RATE_LIMIT_ERROR_DETECTED_IN_EVALUATE_FUNCTION
                )

            score_results.append(
                score_result.ScoreResult(
                    name=metric.name,
                    value=0.0,
                    reason=str(exception),
                    scoring_failed=True,
                )
            )

    return score_results


def _has_evaluation_span_parameter(func: Callable) -> bool:
    try:
        sig = inspect.signature(func)
        has_param = EVALUATION_SPAN_PARAMETER_NAME in sig.parameters
    except (ValueError, TypeError):
        # If we can't inspect the signature, assume no parameter
        has_param = False

    return has_param



"""Upload command for Opik CLI."""

import json
import re
import sys
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, Optional

import click
from rich.console import Console
from rich.progress import Progress, SpinnerColumn, TextColumn

import opik
from opik.rest_api.core.api_error import ApiError
from opik.api_objects.trace import trace_data
from opik.api_objects.span import span_data
from opik.api_objects.trace.migration import prepare_traces_and_spans_for_copy

console = Console()


def _matches_name_pattern(name: str, pattern: Optional[str]) -> bool:
    """Check if a name matches the given regex pattern."""
    if pattern is None:
        return True
    try:
        return bool(re.search(pattern, name))
    except re.error as e:
        console.print(f"[red]Invalid regex pattern '{pattern}': {e}[/red]")
        return False


def _json_to_trace_data(
    trace_info: Dict[str, Any], project_name: str
) -> trace_data.TraceData:
    """Convert JSON trace data to TraceData object."""
    return trace_data.TraceData(
        id=trace_info.get("id", ""),
        name=trace_info.get("name"),
        start_time=(
            datetime.fromisoformat(trace_info["start_time"].replace("Z", "+00:00"))
            if trace_info.get("start_time")
            else None
        ),
        end_time=(
            datetime.fromisoformat(trace_info["end_time"].replace("Z", "+00:00"))
            if trace_info.get("end_time")
            else None
        ),
        metadata=trace_info.get("metadata"),
        input=trace_info.get("input"),
        output=trace_info.get("output"),
        tags=trace_info.get("tags"),
        feedback_scores=trace_info.get("feedback_scores"),
        project_name=project_name,
        created_by=trace_info.get("created_by"),
        error_info=trace_info.get("error_info"),
        thread_id=trace_info.get("thread_id"),
    )


def _json_to_span_data(
    span_info: Dict[str, Any], project_name: str
) -> span_data.SpanData:
    """Convert JSON span data to SpanData object."""
    return span_data.SpanData(
        trace_id=span_info.get("trace_id", ""),
        id=span_info.get("id", ""),
        parent_span_id=span_info.get("parent_span_id"),
        name=span_info.get("name"),
        type=span_info.get("type", "general"),
        start_time=(
            datetime.fromisoformat(span_info["start_time"].replace("Z", "+00:00"))
            if span_info.get("start_time")
            else None
        ),
        end_time=(
            datetime.fromisoformat(span_info["end_time"].replace("Z", "+00:00"))
            if span_info.get("end_time")
            else None
        ),
        metadata=span_info.get("metadata"),
        input=span_info.get("input"),
        output=span_info.get("output"),
        tags=span_info.get("tags"),
        usage=span_info.get("usage"),
        feedback_scores=span_info.get("feedback_scores"),
        project_name=project_name,
        model=span_info.get("model"),
        provider=span_info.get("provider"),
        error_info=span_info.get("error_info"),
        total_cost=span_info.get("total_cost"),
    )


def _import_traces(
    client: opik.Opik,
    project_dir: Path,
    dry_run: bool,
    name_pattern: Optional[str] = None,
) -> int:
    """Import traces from JSON files."""
    trace_files = list(project_dir.glob("trace_*.json"))

    if not trace_files:
        console.print(f"[yellow]No trace files found in {project_dir}[/yellow]")
        return 0

    imported_count = 0
    with Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        console=console,
    ) as progress:
        task = progress.add_task("Uploading traces...", total=len(trace_files))

        for trace_file in trace_files:
            try:
                with open(trace_file, "r", encoding="utf-8") as f:
                    trace_data = json.load(f)

                # Filter by name pattern if specified
                trace_name = trace_data.get("trace", {}).get("name", "")
                if name_pattern and not _matches_name_pattern(trace_name, name_pattern):
                    continue

                if dry_run:
                    print(f"Would upload trace: {trace_data['trace']['id']}")
                    imported_count += 1
                    progress.update(
                        task,
                        description=f"Imported {imported_count}/{len(trace_files)} traces",
                    )
                    continue

                # Extract trace information
                trace_info = trace_data["trace"]
                spans_info = trace_data.get("spans", [])

                # Convert JSON data to TraceData and SpanData objects
                # Use a temporary project name for the migration logic
                temp_project_name = "temp_import"
                trace_data_obj = _json_to_trace_data(trace_info, temp_project_name)

                # Convert spans to SpanData objects, setting the correct trace_id
                span_data_objects = []
                for span_info in spans_info:
                    span_info["trace_id"] = trace_data_obj.id  # Ensure trace_id is set
                    span_data_obj = _json_to_span_data(span_info, temp_project_name)
                    span_data_objects.append(span_data_obj)

                # Use the migration logic to prepare traces and spans with new IDs
                # This handles orphan spans, validates parent relationships, and logs issues
                new_trace_data, new_span_data = prepare_traces_and_spans_for_copy(
                    destination_project_name=client.project_name or "default",
                    traces_data=[trace_data_obj],
                    spans_data=span_data_objects,
                )

                # Create the trace using the prepared data
                new_trace = new_trace_data[0]
                trace_obj = client.trace(
                    name=new_trace.name,
                    start_time=new_trace.start_time,
                    end_time=new_trace.end_time,
                    input=new_trace.input,
                    output=new_trace.output,
                    metadata=new_trace.metadata,
                    tags=new_trace.tags,
                    thread_id=new_trace.thread_id,
                    error_info=new_trace.error_info,
                )

                # Create spans using the prepared data
                for span_data_obj in new_span_data:
                    client.span(
                        trace_id=trace_obj.id,
                        parent_span_id=span_data_obj.parent_span_id,
                        name=span_data_obj.name,
                        type=span_data_obj.type,
                        start_time=span_data_obj.start_time,
                        end_time=span_data_obj.end_time,
                        input=span_data_obj.input,
                        output=span_data_obj.output,
                        metadata=span_data_obj.metadata,
                        tags=span_data_obj.tags,
                        usage=span_data_obj.usage,
                        model=span_data_obj.model,
                        provider=span_data_obj.provider,
                        error_info=span_data_obj.error_info,
                    )

                imported_count += 1
                progress.update(
                    task,
                    description=f"Imported {imported_count}/{len(trace_files)} traces",
                )

            except Exception as e:
                console.print(
                    f"[red]Error importing trace from {trace_file.name}: {e}[/red]"
                )
                continue

    return imported_count


def _import_datasets(
    client: opik.Opik,
    project_dir: Path,
    dry_run: bool,
    name_pattern: Optional[str] = None,
) -> int:
    """Import datasets from JSON files."""
    dataset_files = list(project_dir.glob("dataset_*.json"))

    if not dataset_files:
        console.print(f"[yellow]No dataset files found in {project_dir}[/yellow]")
        return 0

    imported_count = 0
    for dataset_file in dataset_files:
        try:
            with open(dataset_file, "r", encoding="utf-8") as f:
                dataset_data = json.load(f)

            # Filter by name pattern if specified
            dataset_name = dataset_data.get("name", "")
            if name_pattern and not _matches_name_pattern(dataset_name, name_pattern):
                continue

            if dry_run:
                print(f"Would upload dataset: {dataset_data['name']}")
                imported_count += 1
                continue

            # Check if dataset already exists
            try:
                client.get_dataset(dataset_data["name"])
                console.print(
                    f"[yellow]Dataset '{dataset_data['name']}' already exists, skipping...[/yellow]"
                )
                imported_count += 1
                continue
            except ApiError as e:
                if e.status_code == 404:
                    # Dataset doesn't exist, create it
                    pass
                else:
                    # Re-raise other API errors (network, auth, etc.)
                    raise

            # Create dataset
            dataset = client.create_dataset(
                name=dataset_data["name"], description=dataset_data.get("description")
            )

            # Insert dataset items
            for item in dataset_data.get("items", []):
                dataset.insert(
                    [
                        {
                            "input": item["input"],
                            "expected_output": item["expected_output"],
                            "metadata": item.get("metadata"),
                        }
                    ]
                )

            imported_count += 1

        except Exception as e:
            console.print(
                f"[red]Error importing dataset from {dataset_file.name}: {e}[/red]"
            )
            continue

    return imported_count


def _import_prompts(
    client: opik.Opik,
    project_dir: Path,
    dry_run: bool,
    name_pattern: Optional[str] = None,
) -> int:
    """Import prompts from JSON files."""
    prompt_files = list(project_dir.glob("prompt_*.json"))

    if not prompt_files:
        console.print(f"[yellow]No prompt files found in {project_dir}[/yellow]")
        return 0

    imported_count = 0
    for prompt_file in prompt_files:
        try:
            with open(prompt_file, "r", encoding="utf-8") as f:
                prompt_data = json.load(f)

            # Filter by name pattern if specified
            prompt_name = prompt_data.get("name", "")
            if name_pattern and not _matches_name_pattern(prompt_name, name_pattern):
                continue

            if dry_run:
                print(f"Would upload prompt: {prompt_data['name']}")
                imported_count += 1
                continue

            # Create prompt
            client.create_prompt(
                name=prompt_data["name"],
                prompt=prompt_data["current_version"]["prompt"],
                metadata=prompt_data["current_version"].get("metadata"),
            )

            imported_count += 1

        except Exception as e:
            console.print(
                f"[red]Error importing prompt from {prompt_file.name}: {e}[/red]"
            )
            continue

    return imported_count


@click.command(name="import")
@click.argument(
    "workspace_folder",
    type=click.Path(file_okay=False, dir_okay=True, readable=True),
)
@click.argument("workspace_name", type=str)
@click.option(
    "--dry-run",
    is_flag=True,
    help="Show what would be imported without actually importing.",
)
@click.option(
    "--all",
    is_flag=True,
    help="Include all data types (traces, datasets, prompts).",
)
@click.option(
    "--include",
    type=click.Choice(["traces", "datasets", "prompts"], case_sensitive=False),
    multiple=True,
    default=["traces"],
    help="Data types to include in upload. Can be specified multiple times. Defaults to traces only.",
)
@click.option(
    "--exclude",
    type=click.Choice(["traces", "datasets", "prompts"], case_sensitive=False),
    multiple=True,
    help="Data types to exclude from upload. Can be specified multiple times.",
)
@click.option(
    "--name",
    type=str,
    help="Filter items by name using Python regex patterns. Matches against trace names, dataset names, or prompt names.",
)
@click.option(
    "--debug",
    is_flag=True,
    help="Enable debug output to show detailed information about the import process.",
)
def import_data(
    workspace_folder: str,
    workspace_name: str,
    dry_run: bool,
    all: bool,
    include: tuple,
    exclude: tuple,
    name: Optional[str],
    debug: bool,
) -> None:
    """
    Upload data from local files to a workspace or workspace/project.

    This command reads data from JSON files in the specified workspace folder
    and imports them to the specified workspace or project.

    Note: Thread metadata is automatically calculated from traces with the same thread_id,
    so threads don't need to be imported separately.

    WORKSPACE_FOLDER: Directory containing JSON files to import.
    WORKSPACE_NAME: Either a workspace name (e.g., "my-workspace") to import to all projects,
                   or workspace/project (e.g., "my-workspace/my-project") to import to a specific project.
    """
    try:
        if debug:
            console.print("[blue]DEBUG: Starting import with parameters:[/blue]")
            console.print(f"[blue]  workspace_folder: {workspace_folder}[/blue]")
            console.print(f"[blue]  workspace_name: {workspace_name}[/blue]")
            console.print(f"[blue]  include: {include}[/blue]")
            console.print(f"[blue]  debug: {debug}[/blue]")

        # Parse workspace/project from the argument
        if "/" in workspace_name:
            workspace, project_name = workspace_name.split("/", 1)
            import_to_specific_project = True
            if debug:
                console.print(
                    f"[blue]DEBUG: Parsed workspace: {workspace}, project: {project_name}[/blue]"
                )
        else:
            # Only workspace specified - upload to all projects
            workspace = workspace_name
            project_name = None
            import_to_specific_project = False
            if debug:
                console.print(f"[blue]DEBUG: Workspace only: {workspace}[/blue]")

        # Initialize Opik client with workspace
        if debug:
            console.print(
                f"[blue]DEBUG: Initializing Opik client with workspace: {workspace}[/blue]"
            )
        client = opik.Opik(workspace=workspace)

        # Use the specified workspace folder directly
        project_dir = Path(workspace_folder)

        # Determine which data types to upload
        if all:
            # If --all is specified, include all data types
            include_set = {"traces", "datasets", "prompts"}
        else:
            include_set = set(item.lower() for item in include)

        exclude_set = set(item.lower() for item in exclude)

        # Apply exclusions
        data_types = include_set - exclude_set

        if not project_dir.exists():
            console.print(f"[red]Error: Directory not found: {project_dir}[/red]")
            console.print("[yellow]Make sure the path is correct.[/yellow]")
            sys.exit(1)

        console.print(f"[green]Uploading data from {project_dir}[/green]")

        if import_to_specific_project:
            console.print(
                f"[blue]Uploading to workspace: {workspace}, project: {project_name}[/blue]"
            )
        else:
            console.print(
                f"[blue]Uploading to workspace: {workspace} (all projects)[/blue]"
            )

        if debug:
            console.print(f"[blue]Data types: {', '.join(sorted(data_types))}[/blue]")

        # Note about workspace vs project-specific data
        project_specific = [dt for dt in data_types if dt in ["traces"]]
        workspace_data = [dt for dt in data_types if dt in ["datasets", "prompts"]]

        if project_specific and workspace_data:
            if import_to_specific_project:
                console.print(
                    f"[yellow]Note: {', '.join(project_specific)} will be imported to project '{project_name}', {', '.join(workspace_data)} belong to workspace '{workspace}'[/yellow]"
                )
            else:
                console.print(
                    f"[yellow]Note: {', '.join(project_specific)} will be imported to all projects, {', '.join(workspace_data)} belong to workspace '{workspace}'[/yellow]"
                )
        elif workspace_data:
            console.print(
                f"[yellow]Note: {', '.join(workspace_data)} belong to workspace '{workspace}'[/yellow]"
            )

        if dry_run:
            console.print("[yellow]Dry run mode - no data will be imported[/yellow]")

        if import_to_specific_project:
            # Upload to specific project
            # Create a new client instance with the specific project name
            assert project_name is not None  # Type narrowing for mypy
            client = opik.Opik(workspace=workspace, project_name=project_name)

            # Upload each data type
            total_imported = 0

            # Upload traces
            if "traces" in data_types:
                if debug:
                    console.print("[blue]Uploading traces...[/blue]")
                traces_imported = _import_traces(client, project_dir, dry_run, name)
                total_imported += traces_imported

            # Upload datasets
            if "datasets" in data_types:
                if debug:
                    console.print("[blue]Uploading datasets...[/blue]")
                datasets_imported = _import_datasets(client, project_dir, dry_run, name)
                total_imported += datasets_imported

            # Upload prompts
            if "prompts" in data_types:
                if debug:
                    console.print("[blue]Uploading prompts...[/blue]")
                prompts_imported = _import_prompts(client, project_dir, dry_run, name)
                total_imported += prompts_imported

            if dry_run:
                console.print(
                    f"[green]Dry run complete: Would import {total_imported} items[/green]"
                )
            else:
                console.print(
                    f"[green]Successfully imported {total_imported} items to project '{project_name}'[/green]"
                )
        else:
            # Upload to all projects in workspace
            # Get all projects in the workspace
            try:
                projects_response = client.rest_client.projects.find_projects()
                projects = projects_response.content or []

                if not projects:
                    console.print(
                        f"[yellow]No projects found in workspace '{workspace}'[/yellow]"
                    )
                    return

                console.print(
                    f"[blue]Found {len(projects)} projects in workspace[/blue]"
                )

                # Upload workspace-level data once (datasets, experiments, prompts)
                total_imported = 0

                # Upload datasets
                if "datasets" in data_types:
                    if debug:
                        console.print("[blue]Uploading datasets...[/blue]")
                    datasets_imported = _import_datasets(
                        client, project_dir, dry_run, name
                    )
                    total_imported += datasets_imported

                # Upload prompts
                if "prompts" in data_types:
                    if debug:
                        console.print("[blue]Uploading prompts...[/blue]")
                    prompts_imported = _import_prompts(
                        client, project_dir, dry_run, name
                    )
                    total_imported += prompts_imported

                # Note: Traces are project-specific and should be imported to a specific project
                # rather than being uploaded to all projects in a workspace
                if "traces" in data_types:
                    console.print(
                        "[yellow]Note: Traces are project-specific. Use workspace/project format to import traces to a specific project.[/yellow]"
                    )

                if dry_run:
                    console.print(
                        f"[green]Dry run complete: Would import {total_imported} items to workspace '{workspace}'[/green]"
                    )
                else:
                    console.print(
                        f"[green]Successfully imported {total_imported} items to workspace '{workspace}'[/green]"
                    )

            except Exception as e:
                console.print(f"[red]Error getting projects from workspace: {e}[/red]")
                sys.exit(1)

    except Exception as e:
        console.print(f"[red]Error: {e}[/red]")
        sys.exit(1)



from typing import Dict, List, TYPE_CHECKING

if TYPE_CHECKING:
    import langchain_core.messages


def convert_to_langchain_messages(
    messages: List[Dict[str, str]],
) -> List["langchain_core.messages.BaseMessage"]:
    import langchain_core.messages

    langchain_messages = []
    for message in messages:
        role = message["role"]
        content = message["content"]

        if role == "system":
            langchain_messages.append(
                langchain_core.messages.SystemMessage(content=content)
            )
        elif role == "user":
            langchain_messages.append(
                langchain_core.messages.HumanMessage(content=content)
            )
        elif role == "assistant":
            langchain_messages.append(
                langchain_core.messages.AIMessage(content=content)
            )

    return langchain_messages
